#include <ButtonConstants.au3>
#include <WindowsConstants.au3>
#include <GUIListBox.au3>
#include <File.au3>
#include <Array.au3>
#include "_ArrayCompare.au3"

Opt("MustDeclareVars", 1)

Global $iListBkColor = 0xFFFFFF
Global $iListSelColor = 0xFFCCFF
Global $iListDataColor[4] = [0x000000, 0xC00000, 0x008000, 0x0000C0]
Global $lvwDiff
Global $aTI[2] = [0, 0]

; First example : Compare strings down at the character level
_CompareData("The cat sat on the mat", "The dog sat on the cat", 0)

; Second example : Compare strings down at the word level - adjust the delimiters as required
_CompareData("The cat sat on the mat", "The dog sat on the cat", 1)

; Third example : Compare files
Local $aFiles = _SelectFiles()
If $aFiles[0] <> "" Then _CompareData($aFiles[0], $aFiles[1], 2)

; Done
Exit

Func _CompareData($sOld, $sNew, $iCompareMode = 0, $sDelimiters = " .\/[]{}()	")
	Local $i, $j, $aRet

	; Arrays for difference functions
	Local $aLeft, $aRight

	Switch $iCompareMode
		Case 0
			; Characters
			$aLeft = StringSplit($sOld, "", 2)
			$aRight = StringSplit($sNew, "", 2)
		Case 1
			; Strings
			$aLeft = _SplitString($sOld, $sDelimiters)
			$aRight = _SplitString($sNew, $sDelimiters)
		Case 2
			_FileReadToArray($sOld, $aLeft)
			_ArrayDelete($aLeft, 0)
			_FileReadToArray($sNew, $aRight)
			_ArrayDelete($aRight, 0)
		Case 3
			; Window lists
			Dim $aLeft[$sOld[0][0]]
			For $i = 1 To $sOld[0][0]
				$aLeft[$i - 1] = $sOld[$i][1]
			Next
			Dim $aRight[$sNew[0][0]]
			For $i = 1 To $sNew[0][0]
				$aRight[$i - 1] = $sNew[$i][1]
			Next
	EndSwitch

	$aRet = _ArrayCompare($aLeft, $aRight)

	Switch $iCompareMode
		Case 0
			MsgBox(0, "Characters", _GetScript(0, $aLeft, $aRet[0], $aRight, $aRet[1])) ; Shows the edit script for the "old" text
		Case 1
			MsgBox(0, "Words", _GetScript(1, $aLeft, $aRet[0], $aRight, $aRet[1])) ; Shows the edit script for the "new" text
		Case 2
			_ShowDiffDialog($aLeft, $aRet[0], $aRight, $aRet[1]) ; Shows the diff dialog with colour-coded differences
	EndSwitch
EndFunc   ;==>_CompareData

Func _GetScript($nSide, $aLeft, $aLeftScript, $aRight, $aRightScript)
	Local $sRet[2], $nState[2], $i, $j, $nCurr, $nThis
	Local $nLeft = UBound($aLeft), $nRight = UBound($aRight)

	; Blank out return values
	$sRet[0] = ""
	$sRet[1] = ""

	$j = 0
	$nCurr = 0
	For $i = 0 To $nLeft - 1
		Switch $aLeftScript[$i]
			Case 0
				; Skip over insertions in the new text
				While $j < $nRight
					If $aRightScript[$j] = 0 Then ExitLoop
					$j += 1
				WEnd

				If $nCurr <> 0 Then $sRet[0] &= "</m>"
				$nCurr = 0
				$j += 1
			Case 1
				If $j < $nRight Then
					If $aRightScript[$j] = 0 Then
						$nThis = 2 ; deleted
					Else
						$nThis = 1 ; changed
					EndIf
				Else
					$nThis = 2 ; deleted
				EndIf

				If $nThis <> $nCurr Then
					If $nCurr <> 0 Then $sRet[0] &= "</m>"
					$sRet[0] &= "<m "
					If $nThis = 1 Then
						; changed
						$sRet[0] &= "type=""c"""
					Else
						; deleted
						$sRet[0] &= "type=""d"""
					EndIf
					$sRet[0] &= ">"
					$nCurr = $nThis
				EndIf
		EndSwitch
		$sRet[0] &= $aLeft[$i]
	Next
	If $nCurr <> 0 Then $sRet[0] &= "</m>"

	$j = 0
	$nCurr = 0
	For $i = 0 To $nRight - 1
		Switch $aRightScript[$i]
			Case 0
				; Skip over deletions in the old text
				While $j < $nLeft
					If $aLeftScript[$j] <> 1 Then ExitLoop
					$j += 1
				WEnd

				If $nCurr <> 0 Then $sRet[1] &= "</m>"
				$nCurr = 0
				$j += 1
			Case 1
				If $j < $nLeft Then
					If $aLeftScript[$j] = 0 Then
						$nThis = 2 ; inserted
					Else
						$nThis = 1 ; changed
					EndIf
				Else
					$nThis = 2 ; inserted
				EndIf

				If $nThis <> $nCurr Then
					If $nCurr <> 0 Then $sRet[1] &= "</m>"
					$sRet[1] &= "<m "
					If $nThis = 1 Then
						; changed
						$sRet[1] &= "type=""c"""
					Else
						; inserted
						$sRet[1] &= "type=""i"""
					EndIf
					$sRet[1] &= ">"
					$nCurr = $nThis
				EndIf
		EndSwitch
		$sRet[1] &= $aRight[$i]
	Next
	If $nCurr <> 0 Then $sRet[1] &= "</m>"

	Return $sRet[$nSide]
EndFunc   ;==>_GetScript

Func _SplitString($sSplit, $sDelimiters)
	Local $aWords[1], $i, $j = 0, $bDelimiter = False, $cChar
	$aWords[0] = ""
	For $i = 1 To StringLen($sSplit)
		$cChar = StringMid($sSplit, $i, 1)
		If StringInStr($sDelimiters, $cChar) Then
			If $i = 1 Or $bDelimiter Then
				$aWords[$j] &= $cChar
			Else
				$j += 1
				ReDim $aWords[$j + 1]
				$aWords[$j] = $cChar
			EndIf
			$bDelimiter = True
		Else
			If $i = 1 Or (Not $bDelimiter) Then
				$aWords[$j] &= $cChar
			Else
				$j += 1
				ReDim $aWords[$j + 1]
				$aWords[$j] = $cChar
			EndIf
			$bDelimiter = False
		EndIf
	Next

	Return $aWords
EndFunc   ;==>_SplitString

Func _SelectFiles()
	Local $aRet[2], $sTemp, $nMsg
	Local $dlgSelectFiles = GUICreate("Select Files", 422, 108)
	Local $txtFile1 = GUICtrlCreateInput("", 8, 8, 377, 21)
	Local $btnBrowse1 = GUICtrlCreateButton("...", 392, 8, 21, 21, 0)
	Local $txtFile2 = GUICtrlCreateInput("", 8, 39, 377, 21)
	Local $btnBrowse2 = GUICtrlCreateButton("...", 392, 40, 21, 21, 0)
	Local $btnCancel = GUICtrlCreateButton("Cancel", 332, 72, 81, 25, 0)
	Local $btnOK = GUICtrlCreateButton("OK", 245, 72, 81, 25, $BS_DEFPUSHBUTTON)
	GUISetState(@SW_SHOW)

	While 1
		$nMsg = GUIGetMsg()
		Switch $nMsg
			Case $GUI_EVENT_CLOSE, $btnCancel
				$aRet[0] = ""
				$aRet[1] = ""
				ExitLoop
			Case $btnOK
				$aRet[0] = GUICtrlRead($txtFile1)
				$aRet[1] = GUICtrlRead($txtFile2)
				ExitLoop
			Case $btnBrowse1
				$sTemp = FileOpenDialog("Select File", "", "Text files (*.txt)|All files (*.*)", 1, GUICtrlRead($txtFile1))
				If @error = 0 Then GUICtrlSetData($txtFile1, $sTemp)
			Case $btnBrowse2
				$sTemp = FileOpenDialog("Select File", "", "Text files (*.txt)|All files (*.*)", 1, GUICtrlRead($txtFile2))
				If @error = 0 Then GUICtrlSetData($txtFile2, $sTemp)
		EndSwitch
	WEnd

	GUIDelete()

	Return $aRet
EndFunc   ;==>_SelectFiles

Func _ShowDiffDialog($aLeft, $aLeftScript, $aRight, $aRightScript)

	Local $iWidth = Int(@DesktopWidth * 0.75)
	Local $iHeight = Int(@DesktopHeight * 0.75)
	Local $iLvwWidth = Int(($iWidth - 32) / 2)

	Local $iMenuHeight = _WinAPI_GetSystemMetrics($SM_CYMENU)

	#Region ### START Koda GUI section ### Form=F:\Docs\Scripts\AutoItDiff.kxf
	Local $frmMain = GUICreate("AutoItDiff", $iWidth, $iHeight, -1, -1, BitOR($WS_MAXIMIZEBOX, $WS_MINIMIZEBOX, $WS_SIZEBOX, $WS_THICKFRAME, $WS_SYSMENU, $WS_CAPTION, $WS_OVERLAPPEDWINDOW, $WS_TILEDWINDOW, $WS_POPUP, $WS_POPUPWINDOW, $WS_GROUP, $WS_TABSTOP, $WS_BORDER))
	$lvwDiff = GUICtrlCreateListView("Old|New", 8, 8, $iWidth - 16, $iHeight - $iMenuHeight - 16)
	GUICtrlSetFont($lvwDiff, 10, 400, 0, "Courier New")
	GUICtrlSendMsg($lvwDiff, 0x101E, 0, $iLvwWidth - 4)
	GUICtrlSendMsg($lvwDiff, 0x101E, 1, $iLvwWidth - 4)
	Local $mnuFile = GUICtrlCreateMenu("&File")
	Local $mnuFileCompareFolders = GUICtrlCreateMenuItem("Compare folders", $mnuFile)
	Local $mnuFileCompareFiles = GUICtrlCreateMenuItem("Compare files", $mnuFile)
	GUICtrlCreateMenuItem("", $mnuFile)
	Local $mnuFileExit = GUICtrlCreateMenuItem("Exit", $mnuFile)
	Local $frmMain_AccelTable[1][2] = [["^q", $mnuFileExit]]
	GUISetAccelerators($frmMain_AccelTable)
	#EndRegion ### END Koda GUI section ###

	Local $i, $j, $nCurr, $nThis, $liOld, $liNew, $nLeft = UBound($aLeft), $nRight = UBound($aRight)
	$j = 0
	For $i = 0 To $nLeft - 1
		Switch $aLeftScript[$i]
			Case 0
				; Skip over insertions in the new text
				While $j < $nRight
					If $aRightScript[$j] <> 1 Then ExitLoop
					_AddListViewItem($lvwDiff, "", $aRight[$j], 3)
					$j += 1
				WEnd

				_AddListViewItem($lvwDiff, $aLeft[$i], $aRight[$j], 0)

				$j += 1
			Case 1
				If $j < $nRight Then
					If $aRightScript[$j] = 0 Then
						$nThis = 2 ; deleted
					Else
						$nThis = 1 ; changed
					EndIf
				Else
					$nThis = 2 ; deleted
				EndIf

				Switch $nThis
					Case 1
						_AddListViewItem($lvwDiff, $aLeft[$i], $aRight[$j], 2)
						$j += 1
					Case 2
						_AddListViewItem($lvwDiff, $aLeft[$i], "", 1)
				EndSwitch
		EndSwitch
	Next
	While $j < $nRight
		_AddListViewItem($lvwDiff, "", $aRight[$j], 3)
		$j += 1
	WEnd

	GUISetState(@SW_SHOW)

	While 1
		Local $nMsg = GUIGetMsg()
		Switch $nMsg
			Case $GUI_EVENT_CLOSE, $mnuFileExit
				Exit
		EndSwitch
	WEnd

EndFunc   ;==>_ShowDiffDialog

Func _AddListViewItem($lvwHandle, $sText1, $sText2, $iItemData)
	Local $lItem = GUICtrlCreateListViewItem($sText1 & "|" & $sText2, $lvwHandle)
	If $iItemData <> 0 Then GUICtrlSetBkColor($lItem, 0xeeeeee)
	GUICtrlSetColor($lItem, $iListDataColor[$iItemData])
EndFunc   ;==>_AddListViewItem